#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from django.test import TestCase
from sklearn import svm
from sklearn.cluster import KMeans
from sklearn import datasets
from sklearn import linear_model
from test_djangolearn.test_app.models import IrisSVMModel, LinearRegressionModel

iris = datasets.load_iris()

boston = datasets.load_boston()


class InheritanceTestCase(TestCase):
    """Test that a model inheriting from SciKitLearnModel works"""
    def setUp(self):
        clf = svm.SVC()
        X, y = iris.data, iris.target
        clf.fit(X, y)
        self.expected_results = clf.predict(X[0:1])

    def test_model_inheritance_basic(self):
        X, y = iris.data, iris.target

        model = IrisSVMModel.objects.create()
        model.train(X, y)
        results = model.evaluate(X[0:1])
        self.assertEquals(self.expected_results, results)



class MultipleObjectsInheritanceTestCase(TestCase):
    """Test that we segregate the storage layer correctly for different models"""
    def setUp(self):
        # For SVN
        clf = svm.SVC()
        X, y = iris.data, iris.target
        clf.fit(X, y)
        self.svm_expected_results = clf.predict(X[0:1])

        # For Clustering
        lr = linear_model.LinearRegression()
        lr.fit(boston.data, boston.target)

        self.linear_expected = lr.predict(boston.data)


    def test_multiple_model_basic(self):
        """ Test that we don't mess up two models"""
        X, y = iris.data, iris.target

        svm_model = IrisSVMModel.objects.create()
        svm_model.train(X, y)

        lr_model = LinearRegressionModel.objects.create()
        lr_model.train(boston.data, boston.target)

        svm_results = svm_model.evaluate(X[0:1])
        self.assertEquals(self.svm_expected_results, svm_results)


        lr_results = lr_model.evaluate(boston.data)
        # Pick one sample at random
        self.assertEquals(self.linear_expected[10], lr_results[10])




class NotMachineLearningModelException(Exception):
    """Raised when we're trying to store a non machine learning model."""
    pass


class FrameworkVersionCollisionException(Exception):
    """Raised when a stored model has different version from expected."""
    pass

class FeaturesVersionCollisionException(Exception):
    """Raised when a stored model has different version from expected."""
    pass


class ModelStorageException(Exception):
    """Raised when a model has a problem in storage."""
    pass


class ModelNotStoredException(ModelStorageException):
    """Raised when there is no stored model."""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from django.test import TestCase
from sklearn import svm
from sklearn.cluster import KMeans
from sklearn import datasets

from test_djangolearn.test_app.models import IrisSVMModel
from djangolearn.models import SciKitLearnModel
from djangolearn.exceptions import FeaturesVersionCollisionException, FrameworkVersionCollisionException

iris = datasets.load_iris()

class FrameworkCollisionTestCase(TestCase):
    """Test that we pickip a framework collision"""
    def setUp(self):
        clf = svm.SVC()
        X, y = iris.data, iris.target
        clf.fit(X, y)
        self.expected_results = clf.predict(X[0:1])

    def test_framework_collision(self):
        X, y = iris.data, iris.target

        model = IrisSVMModel.objects.create()
        model.train(X, y)

        model.scikit_version = '0.0000'
        model.save()

        with self.assertRaises(FrameworkVersionCollisionException):
            results = model.evaluate(X[0:1])


class DataFrameworkCollisionTestCase(TestCase):
    """Test that we pickip a framework collision"""
    def setUp(self):
        self.clf = svm.SVC()
        X, y = iris.data, iris.target
        self.clf.fit(X, y)
        self.expected_results = self.clf.predict(X[0:1])

    def test_data_framework_collision(self):
        scikit_model = SciKitLearnModel.objects.create()
        scikit_model.store(self.clf, features_version='1')

        with self.assertRaises(FeaturesVersionCollisionException):
            restored = scikit_model.restore(features_version='3')


    def test_data_framework_bypass_one(self):

        scikit_model = SciKitLearnModel.objects.create()
        # We should bypass validation because we did not provide a version
        scikit_model.store(self.clf)

        try:
            restored = scikit_model.restore(features_version='3')
        except (FeaturesVersionCollisionException):
            self.fail()


    def test_data_framework_bypass_two(self):

        scikit_model = SciKitLearnModel.objects.create()
        scikit_model.store(self.clf, features_version='1')

        try:
            # We should bypass validation because we did not provide a version
            restored = scikit_model.restore()
        except (FeaturesVersionCollisionException):
            self.fail()



#!/usr/bin/env python
# encoding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
from os.path import abspath, dirname

import nose


def run_all(argv=None):
    sys.exitfunc = lambda: sys.stderr.write('Shutting down....\n')

    # always insert coverage when running tests through setup.py
    if argv is None:
        argv = [
            'nosetests', '--with-coverage', '--cover-package=djangolearn',
            '--cover-erase', '--verbose',
        ]

    nose.run_exit(
        argv=argv,
        defaultTest=abspath(dirname(__file__))
    )

if __name__ == '__main__':
    run_all(sys.argv)



# encoding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

from django.test import TestCase
from sklearn import svm
from sklearn import datasets

from djangolearn.models import SciKitLearnModel

class SVMBaseTestCase(TestCase):
    def setUp(self):
        self.clf = svm.SVC()
        iris = datasets.load_iris()
        self.X, self.y = iris.data, iris.target
        self.clf.fit(self.X, self.y)
        self.expected_results = self.clf.predict(self.X[0:1])

    def test_model_storage_init(self):
        scikit_model = SciKitLearnModel.objects.create()
        scikit_model.store(self.clf)

        restored = scikit_model.restore()
        restored_results = restored.predict(self.X[0:1])
        self.assertEquals(self.expected_results, restored_results)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pickle
import sklearn
import logging
import shutil
try:
    from tempfile import TemporaryDirectory
except:
    from .py3_utils import TemporaryDirectory

from django.core.files.storage import default_storage

from django.db import models
from django.core.files import File
from django.conf import settings

from djangolearn.exceptions import (
    NotMachineLearningModelException, FrameworkVersionCollisionException,
    FeaturesVersionCollisionException, ModelNotStoredException)

from sklearn.base import BaseEstimator
from sklearn.externals import joblib

logger = logging.getLogger(__name__)

storage_method = getattr(settings, "DJANGOLEARN_STORAGE", None)

if not storage_method:
    logger.debug("DjangoLearn using default django storage method!")
    storage_method = default_storage
else:
    logger.debug("DjangoLearn using custom django storage method %s!" % storage_method)
    storage_method = __import__(storage_method, fromlist=[''])


def model_storage_path(instance, filename):
    # file will be uploaded to MEDIA_ROOT/model/ModelID
    return 'djangolearn/{0}/v{1}__{2}'.format(
            instance.model_handle.id, instance.version, filename)

class MachineLearningModelFileStorage(models.Model):
    """
    A model storage is the method of storing the machine learning model.
    It uses Django FileField for this purpuse, and leverages django-storages
    to indicate to where should the model be saved. Eg. S3/Azure/google
    """

    # File identifier (eg. file name)
    identifier = models.CharField(
        max_length=128, blank=False, null=False)

    # Is header (SciKitLearn uses a file as a header and sequencial numpy
    # arrays. The header has to be specified as the entrypoint)
    is_header = models.BooleanField(default=False)

    # Is active is used in case a model is retrained, and the storage updated.
    active = models.BooleanField(default=True)

    # Is active is used in case a model is retrained, and the storage updated.
    # Version is here to fetch previous versions of the trained model
    version = models.IntegerField(default=1)

    # The handle of the machine learning model the storage belongs to
    model_handle = models.ForeignKey(
        'MachineLearningModel', null=False, related_name='storage_set')

    # The underlying framework version of the model
    framework_version = models.CharField(
        max_length=128, blank=False, null=False)

    # The version of the features framework used to get features
    features_version = models.CharField(
        max_length=128, blank=False, null=True)

    # The file containg the binaries.
    payload = models.FileField(
        upload_to=model_storage_path,
        storage=storage_method)

    def get_payload(self):
        return self.payload.read()


class MachineLearningModelSerialiser(object):
    """A serialiser that takes a machine learning model and sends to storage"""

    def __init__(self, model_object, framework_version, features_version, *args, **kwargs):
        self.model_object = model_object
        self.framework_version = framework_version
        self.features_version = features_version

    def store(self, model, *args, **kwargs):
        raise NotImplementedError("")

    def restore(self, model, *args, **kwargs):
        raise NotImplementedError("")


class ScikitJobLibModelSerialiser(MachineLearningModelSerialiser):
    """
    A serialiser that takes a SciKitLearn model and serialises it via
    JobLib (since Pickle is very insecure). It then uses storage_method
    to persistently store the binaries.
    """

    storage_model = MachineLearningModelFileStorage

    def __init__(self, model_object, *args, **kwargs):
        super(ScikitJobLibModelSerialiser, self).__init__(model_object, *args, **kwargs)

    def store(self, trained_model):

        # hardcoded for now
        model_file_name = 'model.pkl'

        assert '/' not in model_file_name


        prev_storage_obj = self.storage_model.objects.filter(
            active=True,
            model_handle=self.model_object,
        )

        if prev_storage_obj:
            version = prev_storage_obj.first().version
            prev_storage_obj.update(active=False)
            logger.debug(
                "Found previously stored model of version %s" %
                version)
            version = version + 1
        else:
            version = 1

        # Create tempdir
        with TemporaryDirectory() as tmp_dir:
            # save model & get list of all files
            file_names = joblib.dump(trained_model, tmp_dir + '/' + model_file_name)
            # upload files

            for file_name in file_names:
                with open(file_name, mode='ab+') as file_content:
                    file_name = file_name.split('/')[-1]
                    if file_name == model_file_name:
                        is_header = True
                    else:
                        is_header = False

                    storage_obj = self.storage_model.objects.create(
                        version=version,
                        identifier=file_name,
                        payload=File(file_content),
                        model_handle=self.model_object,
                        is_header=is_header,
                        framework_version=self.framework_version,
                        features_version=self.features_version,
                    )


    def restore(self, version=None):
        # hardcoded for now
        model_file_name = 'model.pkl'
        assert '/' not in model_file_name

        # get all files
        if version:
            files = self.storage_model.objects.filter(
                model_handle=self.model_object,
                version=version)
        else:
            files = self.storage_model.objects.filter(
                model_handle=self.model_object,
                active=True)

        if not files:
            raise ModelNotStoredException(
                "No model stored for %s!" % self.model_object)
        else:
            f = files.first()
            # Warning - We will only compare if these arguments were provided
            # If not, validation will be bypassed
            if f.framework_version and self.framework_version and f.framework_version != self.framework_version:
                raise FrameworkVersionCollisionException(
                    "Cannot load a model from a different framework version. "
                    "please retrain the model!")

            elif f.features_version and self.features_version and f.features_version != self.features_version:
                raise FeaturesVersionCollisionException(
                    "Cannot (Or should not) load a model trained with "
                    "different features."
                )

        # Create tempdir
        with TemporaryDirectory() as tmp_dir:

            for storage_file in files:

                logger.debug("Reconstucting %s" % tmp_dir+'/'+storage_file.identifier)
                fs_file = open(tmp_dir+'/'+storage_file.identifier, mode='ab+')
                fs_file.write(storage_file.get_payload())
                fs_file.seek(0)

            # reconstruct
            clf = joblib.load(tmp_dir + '/' + model_file_name)

        return clf


class MachineLearningModel(models.Model):
    # This is a meta class.
    serialiser = MachineLearningModelSerialiser
    loaded_model = None

    model_class_name = models.CharField(
        max_length=128, blank=False, null=False)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    class Meta:
        get_latest_by = "create_time"
        ordering = ('-create_time',)

    def store(self, model, features_version=None, framework_version=None, *args, **kwargs):

        if not self.id:
            try:
                self.refresh_from_db()
            except:
                # compatibility with Django < 1.8
                self.save()

        # Lets double check:
        if not self.id:
            raise Exception(
                "Cannot store a model without an ID.",
                "please reload the model")

        serialiser_instance = self.serialiser(
            self,
            features_version=features_version,
            framework_version=framework_version)
        serialiser_instance.store(model)

    def restore(self, features_version=None, framework_version=None, version=None):
        serialiser_instance = self.serialiser(
            self,
            features_version=features_version,
            framework_version=framework_version)
        self.loaded_model = serialiser_instance.restore(version=version)
        return self.loaded_model


class SciKitLearnModel(MachineLearningModel):
    """A SciKitLearn Machine learning model persistance model"""

    serialiser = ScikitJobLibModelSerialiser

    def __init__(self, *args, **kwargs):
        super(SciKitLearnModel, self).__init__(*args, **kwargs)
        self.scikit_version = sklearn.__version__

    def store(self, model, *args, **kwargs):
        logger.info("Storing Model %s..." % self)

        if not issubclass(model.__class__, BaseEstimator):
            raise NotMachineLearningModelException(
                "Machine Learning Model is not a SciKitLearn Model")

        super(SciKitLearnModel, self).store(
            model, framework_version=self.scikit_version, *args, **kwargs)

    def restore(self, *args, **kwargs):
        logger.info("Restoring Model %s..." % self)

        result = super(SciKitLearnModel, self).restore(
            framework_version=self.scikit_version,
            *args, **kwargs)

        if not issubclass(result.__class__, BaseEstimator):
            raise NotMachineLearningModelException(
                "Restored model is not a SciKitLearn Model")

        return result

    def __str__(self):
        return "SciKit Learn Machine Learning Model %s" % (
                self.id
            )



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import os
from tempfile import mkdtemp



SECRET_KEY = "Please do not spew DeprecationWarnings"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'djangosearch_tests.db',
    }
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',

    'djangolearn',
    'test_djangolearn.test_app',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
            ]
        },
    },
]

MIDDLEWARE_CLASSES = [
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]



# encoding: utf-8
from __future__ import absolute_import

import os

test_runner = None
old_config = None

os.environ['DJANGO_SETTINGS_MODULE'] = 'test_djangolearn.settings'


import django
django.setup()


def setup():
    global test_runner
    global old_config

    from django.test.runner import DiscoverRunner

    test_runner = DiscoverRunner()
    test_runner.setup_test_environment()
    old_config = test_runner.setup_databases()


def teardown():
    test_runner.teardown_databases(old_config)
    test_runner.teardown_test_environment()



# -*- coding: utf-8 -*-

from sklearn import svm
from djangolearn.models import SciKitLearnModel
from sklearn import datasets
from sklearn import linear_model



class IrisSVMModel(SciKitLearnModel):

    def train(self, x, y):
        clf = svm.SVC()
        clf.fit(x, y)
        self.store(clf)

    def evaluate(self, x):
        # this is bad don't load the model at every run
        restored = self.restore()
        restored_results = restored.predict(x)
        return restored_results


class LinearRegressionModel(SciKitLearnModel):

    def train(self, x, y):
        lr = linear_model.LinearRegression()
        lr.fit(x, y)
        self.store(lr)

    def evaluate(self, x):
        # this is bad don't load the model at every run
        restored = self.restore()
        restored_results = restored.predict(x)
        return restored_results



#!/usr/bin/python
# -*- coding: utf-8 -*-

import versioneer
from pip.req import parse_requirements

import subprocess
from datetime import datetime

from setuptools import setup, find_packages

# Parses pip requirements and transform this
# into an array to be used at setup time
install_reqs = [str(req.req) for req in
                parse_requirements("./requirements.txt", session={})]

test_reqs = [str(req.req) for req in
                parse_requirements("./test_requirements.txt", session={})]


tests_require = install_reqs + test_reqs

setup(
    name='djangolearn',
    version = versioneer.get_version(),
    cmdclass = versioneer.get_cmdclass(),
    author='Sergio',
    author_email='sergio@holvi.com',
    packages=find_packages(exclude=('test_djangolearn')),
    install_requires=install_reqs,
    zip_safe=False,
    include_package_data=True,
    test_suite="test_djangolearn.run_tests.run_all",
    tests_require=tests_require,
    setup_requires=['setuptools_scm'],
)
